import csv
import pytest

from pathlib import Path
from typing import Any
from pytest_bdd import scenarios, given, when, then
from nd2k.main import main


DataTable = list[list[str]]


scenarios("./novadax_to_koinly.feature")


@given("a NovaDAX CSV file has the following operations:")
def create_input_file(tmp_path: Path, datatable: DataTable) -> None:
	file = tmp_path / "temp.csv"
	with file.open("w") as f:
		writer = csv.writer(f)
		writer.writerows(datatable)


@when("the file is processed")
def invoke_nd2k_tool(tmp_path: Path, monkeypatch: Any) -> None:
	file = tmp_path / "temp.csv"
	monkeypatch.setattr("sys.argv", ["nd2k", str(file)])
	main()


@when("I attempt to process the file", target_fixture="error_msg")
def invoke_nd2k_tool_with_bad_file(tmp_path: Path, capsys: Any, monkeypatch: Any) -> str:
	file = tmp_path / "temp.csv"
	monkeypatch.setattr("sys.argv", ["nd2k", str(file)])
	with pytest.raises(SystemExit) as exc_info:
		main()
	assert exc_info.value.code == 1
	return str(capsys.readouterr().out)


@then("a Koinly universal file should be created with the following transactions:")
def compare_gerenared_file(tmp_path: Path, datatable: DataTable) -> None:
	file = tmp_path / "temp_koinly_universal.csv"
	with file.open("r") as f:
		result = list(csv.reader(f))
	assert datatable[1:] == result[1:] # ignoring headers


@then("the following error should appear:")
def compare_error_message(error_msg: str, docstring: str) -> None:
	assert error_msg.strip() == docstring.strip()



from collections import defaultdict
from dataclasses import dataclass
from datetime import datetime
from decimal import Decimal
from enum import Enum
from typing import NamedTuple


CSV = list[list[str]]


class OperationType(Enum):
	CRYPTO_DEPOSIT  = "Depósito de criptomoedas"
	FIAT_DEPOSIT    = "Depósito em Reais"
	CRYPTO_WITHDRAW = "Saque de criptomoedas"
	FIAT_WITHDRAW   = "Saque em Reais"
	WITHDRAW_FEE    = "Taxa de saque de criptomoedas"
	REDEEMED_BONUS  = "Redeemed Bonus"
	BUY             = "Compra"
	SELL            = "Venda"
	TRADING_FEE     = "Taxa de transação"


class KoinlyTag(Enum):
	CRYPTO_DEPOSIT  = "deposit"
	FIAT_DEPOSIT    = "deposit"
	CRYPTO_WITHDRAW = "withdraw"
	FIAT_WITHDRAW   = "withdraw"
	WITHDRAW_FEE    = "fee"
	REDEEMED_BONUS  = "reward"
	BUY             = "trade"
	SELL            = "trade"
	TRADING_FEE     = "fee"


@dataclass
class Operation:
	date:    datetime
	type:    OperationType
	summary: str
	symbol:  str
	amount:  Decimal
	status:  str


@dataclass
class NonTrade: # a.k.a. "Simple Transaction"
	operation: Operation

	@property
	def date(self) -> datetime:
		return self.operation.date

	@property
	def summary(self) -> str:
		return self.operation.summary


class TradingPair(NamedTuple):
	"""
	In a trading pair like BTC/EUR:
	- Base asset (BTC): The asset being bought or sold.
	- Quote asset (EUR): The asset used to price the base asset.
	"""
	base:  str
	quote: str


@dataclass
class Trade:
	summary:      str
	trading_pair: TradingPair
	base_asset:   Operation
	quote_asset:  Operation
	trading_fee:  Operation

	@property
	def date(self) -> datetime:
		return self.base_asset.date

	@property
	def type(self) -> OperationType:
		return self.base_asset.type


Transaction = Trade | NonTrade
TransactionGroups = defaultdict[str, list[Transaction]]


@dataclass
class PartialTrade:
	summary:      str
	trading_pair: TradingPair
	base_asset:   Operation | None = None
	quote_asset:  Operation | None = None
	trading_fee:  Operation | None = None

	@property
	def type(self) -> OperationType:
		return self._any_asset.type

	@property
	def _any_asset(self) -> Operation:
		any_asset = self.base_asset or self.quote_asset
		if any_asset:
			return any_asset
		raise ValueError("Empty Trade")

	def complete(self) -> Trade:
		return Trade(**vars(self))



from pathlib import Path
from nd2k.main import read

# https://github.com/thiagoalessio/nd2k/issues/8
def test_handle_input_with_any_encoding(tmp_path: Path) -> None:
	file = tmp_path / "temp.csv"

	contents = (
		bytes("hea,ders,x\r", "ascii")
		+ "foo,".encode("latin-1")
		+ b"\x81\xff\xfe"
		+ "bar,".encode("cp1252")
		+ b"\x80"
		+ "test".encode("utf-8"))

	with file.open("wb") as f:
		f.write(contents)

	assert read(str(file)) == [["foo", "bar", "test"]]



import re
import sys
import os
import csv

from . import __version__, types, converter, formatter
from .types import Transaction


def main() -> None:
	if len(sys.argv) != 2:
		print("Usage: nd2k <novadax-csv>")
		exit(1)

	if sys.argv[1] in ["-v", "--version"]:
		print(__version__)
		exit(0)

	input_file = sys.argv[1]

	if not os.path.exists(input_file):
		print(f"Error: No such file: {input_file}")
		exit(1)

	csv_rows     = read(input_file)
	transactions = converter.organize_rows(csv_rows)
	combined     = converter.combine_by_timestamp(transactions)
	ordered      = order_by_date(combined)
	formatted    = formatter.universal(ordered)

	output_file = re.sub(r"\.csv$", "", input_file) + "_koinly_universal.csv"
	write(output_file, formatted)


def read(path: str) -> types.CSV:
	"""
	The NovaDAX CSV file must be read in reverse order to
	correctly match trades with their corresponding fees.
	"""
	with open(path, "r", encoding="utf-8", errors="ignore") as f:
		return list(reversed(list(csv.reader(f))[1:]))


def order_by_date(lst: list[Transaction]) -> list[Transaction]:
	return sorted(lst, key=lambda t: t.date)


def write(path: str, contents: types.CSV) -> None:
	with open(path, "w", encoding="utf-8", newline="\n") as f:
		writer = csv.writer(f)
		writer.writerows(contents)



from collections import defaultdict
from decimal import Decimal
from typing import cast

from .organize_rows import create_partial_trade
from ..types import TransactionGroups, Transaction, Trade, NonTrade


def combine_by_timestamp(lst: list[Transaction]) -> list[Transaction]:
	return combine_groups(group_by_timestamp(lst))


def group_by_timestamp(lst: list[Transaction]) -> TransactionGroups:
	"""
	Group transactions that have the same timestamp+summary
	"""
	groups: TransactionGroups = defaultdict(list)
	for t in lst:
		groups[generate_group_index(t)].append(t)
	return groups


def generate_group_index(t: Transaction) -> str:
	if isinstance(t, Trade):
		return str(t.date) + t.summary
	else:
		return str(t.date) + t.summary + t.operation.symbol


def combine_groups(groups: TransactionGroups) -> list[Transaction]:
	"""
	Combine all transactions of a group into
	a single transaction, summing up their amounts.
	"""
	combined: list[Transaction] = []

	for grp in groups.values():
		if all(isinstance(i, Trade) for i in grp):
			combined.append(combine_trades(cast(list[Trade], grp)))
		else:
			combined.append(combine_non_trades(cast(list[NonTrade], grp)))

	return combined


def combine_trades(lst: list[Trade]) -> Trade:
	pt = create_partial_trade(lst[0].base_asset)
	pt.quote_asset = lst[0].quote_asset
	pt.trading_fee = lst[0].trading_fee
	tr = pt.complete()

	tr.base_asset.amount  = Decimal(sum(i.base_asset.amount  for i in lst))
	tr.quote_asset.amount = Decimal(sum(i.quote_asset.amount for i in lst))
	tr.trading_fee.amount = Decimal(sum(i.trading_fee.amount for i in lst))

	return tr

def combine_non_trades(lst: list[NonTrade]) -> NonTrade:
	op = lst[0].operation
	op.amount = Decimal(sum(i.operation.amount for i in lst))
	return NonTrade(operation=op)



from datetime import datetime
from .queries import is_a_purchase, is_sending_funds
from .types import Transaction, CSV, Trade, NonTrade, KoinlyTag


universal_csv_headers = [
	"Date",
	"Sent Amount",
	"Sent Currency",
	"Received Amount",
	"Received Currency",
	"Fee Amount",
	"Fee Currency",
	"Net Worth Amount",
	"Net Worth Currency",
	"Label",
	"Description",
]


def universal(transactions: list[Transaction]) -> CSV:
	rows = []
	rows.append(universal_csv_headers)

	for t in transactions:
		if isinstance(t, Trade):
			rows.append(format_trade(t))
		else:
			rows.append(format_non_trade(t))

	return rows


def format_trade(t: Trade) -> list[str]:
	if is_a_purchase(t):
		sent_amount = t.quote_asset.amount
		sent_symbol = t.quote_asset.symbol
		recv_amount = t.base_asset.amount
		recv_symbol = t.base_asset.symbol
	else:
		sent_amount = t.base_asset.amount
		sent_symbol = t.base_asset.symbol
		recv_amount = t.quote_asset.amount
		recv_symbol = t.quote_asset.symbol

	return [
		format_date(t.base_asset.date), # Date
		f"{sent_amount}",               # Sent Amount
		f"{sent_symbol}",               # Sent Currency
		f"{recv_amount}",               # Received Amount
		f"{recv_symbol}",               # Received Currency
		f"{t.trading_fee.amount}",      # Fee Amount
		f"{t.trading_fee.symbol}",      # Fee Currency
		"",                             # Net Worth Amount
		"",                             # Net Worth Currency
		KoinlyTag[t.type.name].value,   # Label
		t.base_asset.summary,           # Description
		"",                             # TxHash
	]


def format_non_trade(t: NonTrade) -> list[str]:
	op = t.operation

	if is_sending_funds(op):
		sent_amount = str(op.amount)
		sent_symbol = op.symbol
		recv_amount = ""
		recv_symbol = ""
	else:
		sent_amount = ""
		sent_symbol = ""
		recv_amount = str(op.amount)
		recv_symbol = op.symbol

	return [
		format_date(op.date),          # Date
		sent_amount,                   # Sent Amount
		sent_symbol,                   # Sent Currency
		recv_amount,                   # Received Amount
		recv_symbol,                   # Received Currency
		"",                            # Fee Amount
		"",                            # Fee Currency
		"",                            # Net Worth Amount
		"",                            # Net Worth Currency
		KoinlyTag[op.type.name].value, # Label
		op.summary,                    # Description
		"",                            # TxHash
	]


def format_date(data: datetime) -> str:
	return data.strftime("%Y-%m-%d %H:%M:%S")



from typing import Any
from datetime import datetime
from decimal import Decimal
from nd2k.types import (
	Operation,
	OperationType,
	PartialTrade,
	TradingPair,
)


def fake_op(**kwargs: Any) -> Operation:
	defaults = {
		"date":    datetime.now(),
		"type":    OperationType.BUY,
		"summary": "Test",
		"symbol":  "TST",
		"amount":  Decimal("1.234"),
		"status":  "Sucesso",
	}
	return Operation(**{**defaults, **kwargs})


def fake_partial_trade(**kwargs: Any) -> PartialTrade:
	defaults = {
		"summary": "BUY(ABC/XYZ)",
		"trading_pair": TradingPair(base="ABC", quote="XYZ"),
	}
	return PartialTrade(**{**defaults, **kwargs})



from datetime import datetime, timedelta
from decimal import Decimal
from typing import cast

from nd2k.converter import combine_by_timestamp
from nd2k.types import OperationType, Trade, TradingPair, NonTrade
from ..helpers import fake_op, fake_partial_trade

def test_combine_by_timestamp() -> None:
	trades     = example_trades()
	non_trades = example_non_trades()
	combined   = combine_by_timestamp(trades + non_trades)

	# t0+t1 | t2 | t3+t4 | t5+t7 | t6 | o0+o1 | o2 | o3 | o4 | o5+o7 | o6
	assert len(combined) == 11

	# t0+t1
	assert cast(Trade, combined[0]).base_asset.amount  == Decimal("30") # 11+19
	assert cast(Trade, combined[0]).quote_asset.amount == Decimal("36") # 13+23
	assert cast(Trade, combined[0]).trading_fee.amount == Decimal("46")  # 17+29

	# t2+t3
	assert cast(Trade, combined[1]).base_asset.amount  == Decimal("74") # 31+43
	assert cast(Trade, combined[1]).quote_asset.amount == Decimal("84") # 37+47
	assert cast(Trade, combined[1]).trading_fee.amount == Decimal("94")  # 41+53

	assert combined[2] == trades[4] # t4

	# t5+t7
	assert cast(Trade, combined[3]).base_asset.amount  == Decimal("172") # 71+101
	assert cast(Trade, combined[3]).quote_asset.amount == Decimal("176") # 73+103
	assert cast(Trade, combined[3]).trading_fee.amount == Decimal("186")  # 79+107

	assert combined[4] == trades[6] # t6

	# o0+o1
	assert cast(NonTrade, combined[5]).operation.amount  == Decimal("5") # 2+3

	assert combined[6] == non_trades[2] # o2
	assert combined[7] == non_trades[3] # o3
	assert combined[8] == non_trades[4] # o4

	# o5+o7
	assert cast(NonTrade, combined[9]).operation.amount  == Decimal("32") # 13+19

	assert combined[10] == non_trades[6] #o6

def example_trades() -> list[Trade]:
	now = datetime.now()
	a_minute_ago = now - timedelta(minutes=1)

	t0 = trade("BUY", "Buy(AAA/USD)", "AAA", "USD", now)
	t0.base_asset.amount  = Decimal("11")
	t0.quote_asset.amount = Decimal("13")
	t0.trading_fee.amount = Decimal("17")

	t1 = trade("BUY", "Buy(AAA/USD)", "AAA", "USD", now)
	t1.base_asset.amount  = Decimal("19")
	t1.quote_asset.amount = Decimal("23")
	t1.trading_fee.amount = Decimal("29")

	t2 = trade("SELL", "Sell(AAA/USD)", "AAA", "USD", now)
	t2.base_asset.amount  = Decimal("31")
	t2.quote_asset.amount = Decimal("37")
	t2.trading_fee.amount = Decimal("41")

	t3 = trade("SELL", "Sell(AAA/USD)", "AAA", "USD", now)
	t3.base_asset.amount  = Decimal("43")
	t3.quote_asset.amount = Decimal("47")
	t3.trading_fee.amount = Decimal("53")

	t4 = trade("BUY", "Buy(AAA/USD)", "AAA", "USD", a_minute_ago)
	t4.base_asset.amount  = Decimal("59")
	t4.quote_asset.amount = Decimal("61")
	t4.trading_fee.amount = Decimal("67")

	t5 = trade("BUY", "Buy(BBB/USD)", "BBB", "USD", now)
	t5.base_asset.amount  = Decimal("71")
	t5.quote_asset.amount = Decimal("73")
	t5.trading_fee.amount = Decimal("79")

	t6 = trade("BUY", "Buy(BBB/USD)", "BBB", "USD", a_minute_ago)
	t6.base_asset.amount  = Decimal("83")
	t6.quote_asset.amount = Decimal("89")
	t6.trading_fee.amount = Decimal("97")

	t7 = trade("BUY", "Buy(BBB/USD)", "BBB", "USD", now)
	t7.base_asset.amount  = Decimal("101")
	t7.quote_asset.amount = Decimal("103")
	t7.trading_fee.amount = Decimal("107")

	return [t for t in [t0, t1, t2, t3, t4, t5, t6, t7]]


def example_non_trades() -> list[NonTrade]:
	now = datetime.now()
	a_minute_ago = now - timedelta(minutes=1)

	o0 = fake_op(summary="withdraw", symbol="AAA", amount=Decimal("2"),  date=now)
	o1 = fake_op(summary="withdraw", symbol="AAA", amount=Decimal("3"),  date=now)
	o2 = fake_op(summary="withdraw", symbol="BBB", amount=Decimal("5"),  date=now)
	o3 = fake_op(summary="withdraw", symbol="AAA", amount=Decimal("7"),  date=a_minute_ago)
	o4 = fake_op(summary="deposit",  symbol="AAA", amount=Decimal("11"), date=now)
	o5 = fake_op(summary="deposit",  symbol="BBB", amount=Decimal("13"), date=now)
	o6 = fake_op(summary="deposit",  symbol="BBB", amount=Decimal("17"), date=a_minute_ago)
	o7 = fake_op(summary="deposit",  symbol="BBB", amount=Decimal("19"), date=now)

	return [NonTrade(operation=o) for o in [o0, o1, o2, o3, o4, o5, o6, o7]]


def trade(ot: str, s: str, b: str, q: str, d: datetime) -> Trade:
	t = fake_partial_trade()

	t.summary = s
	t.trading_pair = TradingPair(b, q)

	t.base_asset = fake_op()
	t.base_asset.date    = d
	t.base_asset.type    = OperationType[ot]
	t.base_asset.summary = s
	t.base_asset.symbol  = b
	t.base_asset.amount  = Decimal("0")

	t.quote_asset = fake_op()
	t.quote_asset.date    = d
	t.quote_asset.type    = OperationType[ot]
	t.quote_asset.summary = s
	t.quote_asset.symbol  = q
	t.quote_asset.amount  = Decimal("0")

	t.trading_fee = fake_op()
	t.trading_fee.date    = d
	t.trading_fee.type    = OperationType.TRADING_FEE
	t.trading_fee.summary = "trading fee"
	t.trading_fee.symbol  = b if ot == "BUY" else q
	t.trading_fee.amount  = Decimal("0")

	return t.complete()



import unicodedata
import re
from datetime import datetime
from decimal import Decimal
from ..types import (
	CSV,
	NonTrade,
	Operation,
	OperationType,
	PartialTrade,
	Trade,
	TradingPair,
	Transaction,
)
from .. import queries as q


def organize_rows(rows: CSV) -> list[Transaction]:
	"""
	Each row in the NovaDAX CSV is an Operation,
	but Koinly expects each row to be a Transaction.
	Transactions can consist of one or more operations:
	- Trades have 3 operations: base asset, quote asset and trading fee.
	- Non-Trades, such as deposits and withdraws, have a single operation.
	"""
	trades: list[Trade] = []
	non_trades: list[NonTrade] = []
	partial_trades: list[PartialTrade] = []

	for row in rows:
		op = create_operation(row)
		if not q.is_successful(op):
			continue

		if not q.is_part_of_a_trade(op):
			non_trades.append(NonTrade(operation=op))
			continue

		tr = create_or_update_trade(op, partial_trades)

		if q.is_completed(tr):
			trades.append(tr.complete())
			partial_trades.remove(tr)

	if len(partial_trades):
		organize_rows_failed(partial_trades)

	return trades + non_trades


def create_or_update_trade(op: Operation, lst: list[PartialTrade]) -> PartialTrade:
	"""
	Attempts to fit operation into an existing partial trade.
	If it doesn't find any match, create a new partial trade.
	"""
	for pt in lst:
		if q.fits_as_base_asset(op, pt):
			pt.base_asset = op
			return pt

		if q.fits_as_quote_asset(op, pt):
			pt.quote_asset = op
			return pt

		if q.fits_as_trading_fee(op, pt):
			pt.trading_fee = op
			return pt

	pt = create_partial_trade(op)
	lst.append(pt)
	return pt


def organize_rows_failed(lst: list[PartialTrade]) -> None:
	"""
	If by the end of "organize_rows" we still have partial trades,
	something is wrong and the program shouldn't proceed.
	"""
	error_msg = "Error! The script went through all rows in the NovaDAX CSV "
	error_msg+= "and could not complete the following trades:\n\n"

	for pt in lst:
		base  = print_operation(pt.base_asset)
		quote = print_operation(pt.quote_asset)
		fee   = print_operation(pt.trading_fee)

		error_msg+= f"base asset:  {base}\n"
		error_msg+= f"quote asset: {quote}\n"
		error_msg+= f"trading fee: {fee}\n\n"

	error_msg+= "The input file may be faulty, "
	error_msg+= "or the script misinterpreted its contents.\n"
	error_msg+= "If you are sure the input file is correct, please open an "
	error_msg+= "issue at https://github.com/thiagoalessio/nd2k/issues/new "
	error_msg+= "and attach the file that caused this error.\n"
	print(error_msg)
	exit(1)


def create_operation(csv_line: list[str]) -> Operation:
	summary = unicodedata.normalize('NFC', csv_line[1])
	return Operation(
		date    = parse_date(csv_line[0]),
		type    = OperationType(summary.split("(")[0]),
		summary = summary,
		symbol  = csv_line[2],
		amount  = parse_amount(csv_line[3]),
		status  = csv_line[4],
	)


def create_partial_trade(op: Operation) -> PartialTrade:
	tr = PartialTrade(
		summary      = op.summary,
		trading_pair = parse_trading_pair(op.summary))
	tr.base_asset  = op if q.fits_as_base_asset(op, tr)  else None
	tr.quote_asset = op if q.fits_as_quote_asset(op, tr) else None
	return tr


def parse_date(data: str) -> datetime:
	return datetime.strptime(data, "%d/%m/%Y %H:%M:%S")


def parse_amount(data: str) -> Decimal:
	pattern = r"^\D*([,\d]+)"
	matches = re.search(pattern, data)

	if not matches:
		raise ValueError(f"No numeric values found in \"{data}\"")

	parts = matches.group(1).split(",")
	last_part = parts.pop()

	# number has no commas, no decimals
	if not len(parts):
		return Decimal(last_part)

	# last comma acting as decimal separator
	int_part = "".join(parts)
	return Decimal(f"{int_part}.{last_part}")


def parse_trading_pair(data: str) -> TradingPair:
	match = re.search(r"\(([^\/]+)\/([^\)]+)", data)
	if match:
		return TradingPair(*match.groups())
	raise ValueError(f"No trading pair found in \"{data}\"")


def print_operation(op: Operation | None) -> str:
	if not op:
		return "<empty>"
	return " | ".join([str(v) for k,v in op.__dict__.items() if k != "type"])



import pytest

from nd2k import queries as q
from nd2k.types import OperationType
from .helpers import fake_op, fake_partial_trade


def test_is_successful() -> None:
	assert q.is_successful(fake_op(status="Sucesso"))
	assert not q.is_successful(fake_op(status="Other"))


def test_is_part_of_a_trade() -> None:
	assert q.is_part_of_a_trade(fake_op(type=OperationType.BUY))
	assert q.is_part_of_a_trade(fake_op(type=OperationType.SELL))
	assert q.is_part_of_a_trade(fake_op(type=OperationType.TRADING_FEE))

	assert not q.is_part_of_a_trade(fake_op(type=OperationType.CRYPTO_DEPOSIT))
	assert not q.is_part_of_a_trade(fake_op(type=OperationType.FIAT_DEPOSIT))
	assert not q.is_part_of_a_trade(fake_op(type=OperationType.CRYPTO_WITHDRAW))
	assert not q.is_part_of_a_trade(fake_op(type=OperationType.FIAT_WITHDRAW))
	assert not q.is_part_of_a_trade(fake_op(type=OperationType.WITHDRAW_FEE))
	assert not q.is_part_of_a_trade(fake_op(type=OperationType.REDEEMED_BONUS))


def test_is_completed() -> None:
	assert q.is_completed(fake_partial_trade(
		base_asset  = fake_op(),
		quote_asset = fake_op(),
		trading_fee = fake_op(),
	))
	assert not q.is_completed(fake_partial_trade(
		base_asset  = fake_op(),
		quote_asset = fake_op(),
		trading_fee = None,
	))
	assert not q.is_completed(fake_partial_trade(
		base_asset  = fake_op(),
		quote_asset = None,
		trading_fee = fake_op(),
	))
	assert not q.is_completed(fake_partial_trade(
		base_asset  = None,
		quote_asset = fake_op(),
		trading_fee = fake_op(),
	))
	assert not q.is_completed(fake_partial_trade(
		base_asset  = fake_op(),
		quote_asset = None,
		trading_fee = None,
	))
	assert not q.is_completed(fake_partial_trade(
		base_asset  = None,
		quote_asset = fake_op(),
		trading_fee = None,
	))
	assert not q.is_completed(fake_partial_trade(
		base_asset  = None,
		quote_asset = None,
		trading_fee = fake_op(),
	))


def test_fits_as_base_asset() -> None:
	tr = fake_partial_trade(
		quote_asset = fake_op(summary="BUY(ABC/XYZ)", symbol="XYZ"),
		trading_fee = fake_op(),
	)
	op = fake_op(summary="BUY(ABC/XYZ)", symbol="ABC")
	assert q.fits_as_base_asset(op, tr)


def test_fits_as_base_asset_already_has_base() -> None:
	tr = fake_partial_trade(
		base_asset = fake_op(summary="BUY(ABC/XYZ)", symbol="XYZ"),
	)
	op = fake_op(summary="BUY(ABC/XYZ)", symbol="ABC")
	assert not q.fits_as_base_asset(op, tr)


def test_fits_as_base_asset_wrong_summary() -> None:
	tr = fake_partial_trade(
		quote_asset = fake_op(summary="BUY(ABC/XYZ)", symbol="XYZ"),
		trading_fee = fake_op(),
	)
	op = fake_op(summary="SELL(ABC/XYZ)", symbol="ABC")
	assert not q.fits_as_base_asset(op, tr)


def test_fits_as_base_asset_wrong_symbol() -> None:
	tr = fake_partial_trade(
		quote_asset = fake_op(summary="BUY(ABC/XYZ)", symbol="XYZ"),
		trading_fee = fake_op(),
	)
	op = fake_op(summary="BUY(ABC/XYZ)", symbol="XYZ")
	assert not q.fits_as_base_asset(op, tr)


def test_fits_as_quote_asset() -> None:
	tr = fake_partial_trade(
		base_asset  = fake_op(summary="BUY(ABC/XYZ)", symbol="ABC"),
		trading_fee = fake_op(),
	)
	op = fake_op(summary="BUY(ABC/XYZ)", symbol="XYZ")
	assert q.fits_as_quote_asset(op, tr)


def test_fits_as_quote_asset_already_has_quote() -> None:
	tr = fake_partial_trade(
		quote_asset = fake_op(summary="BUY(ABC/XYZ)", symbol="XYZ"),
	)
	op = fake_op(summary="BUY(ABC/XYZ)", symbol="XYZ")
	assert not q.fits_as_quote_asset(op, tr)


def test_fits_as_quote_asset_wrong_summary() -> None:
	tr = fake_partial_trade(
		base_asset  = fake_op(summary="BUY(ABC/XYZ)", symbol="ABC"),
		trading_fee = fake_op(),
	)
	op = fake_op(summary="SELL(ABC/XYZ)", symbol="XYZ")
	assert not q.fits_as_quote_asset(op, tr)


def test_fits_as_quote_asset_wrong_symbol() -> None:
	tr = fake_partial_trade(
		base_asset  = fake_op(summary="BUY(ABC/XYZ)", symbol="ABC"),
		trading_fee = fake_op(),
	)
	op = fake_op(summary="BUY(ABC/XYZ)", symbol="ABC")
	assert not q.fits_as_quote_asset(op, tr)


def test_fits_as_trading_fee_buy() -> None:
	tr = fake_partial_trade(
		base_asset  = fake_op(symbol="ABC", type=OperationType.BUY),
		quote_asset = fake_op(symbol="XYZ", type=OperationType.BUY),
	)
	fee = fake_op(symbol="ABC", type=OperationType.TRADING_FEE)
	assert q.fits_as_trading_fee(fee, tr)


def test_fits_as_trading_fee_sell() -> None:
	tr = fake_partial_trade(
		base_asset  = fake_op(symbol="ABC", type=OperationType.SELL),
		quote_asset = fake_op(symbol="XYZ", type=OperationType.SELL),
	)
	fee = fake_op(symbol="XYZ", type=OperationType.TRADING_FEE)
	assert q.fits_as_trading_fee(fee, tr)


def test_fits_as_trading_fee_already_has_fee() -> None:
	tr = fake_partial_trade(
		base_asset  = fake_op(),
		quote_asset = fake_op(),
		trading_fee = fake_op(),
	)
	fee = fake_op(type=OperationType.TRADING_FEE)
	assert not q.fits_as_trading_fee(fee, tr)


def test_fits_as_trading_fee_sell_wrong_type() -> None:
	tr = fake_partial_trade()
	op = fake_op(type=OperationType.WITHDRAW_FEE)
	assert not q.fits_as_trading_fee(op, tr)


def test_fits_as_trading_fee_buy_wrong_symbol() -> None:
	tr = fake_partial_trade(
		base_asset  = fake_op(symbol="ABC", type=OperationType.BUY),
		quote_asset = fake_op(symbol="XYZ", type=OperationType.BUY),
	)
	fee = fake_op(symbol="XYZ", type=OperationType.TRADING_FEE)
	assert not q.fits_as_trading_fee(fee, tr)


def test_fits_as_trading_fee_sell_wrong_symbol() -> None:
	tr = fake_partial_trade(
		base_asset  = fake_op(symbol="ABC", type=OperationType.SELL),
		quote_asset = fake_op(symbol="XYZ", type=OperationType.SELL),
	)
	fee = fake_op(symbol="ABC", type=OperationType.TRADING_FEE)
	assert not q.fits_as_trading_fee(fee, tr)


def test_fits_as_trading_fee_empty_trade() -> None:
	tr = fake_partial_trade()
	op = fake_op(type=OperationType.TRADING_FEE)
	with pytest.raises(ValueError) as e:
		q.fits_as_trading_fee(op, tr)
	assert str(e.value) == "Empty Trade"


def test_fits_as_trading_fee_malformed_trade() -> None:
	tr = fake_partial_trade(
		base_asset = fake_op(type=OperationType.CRYPTO_DEPOSIT),
	)
	op = fake_op(type=OperationType.TRADING_FEE)
	with pytest.raises(ValueError) as e:
		q.fits_as_trading_fee(op, tr)
	assert str(e.value) == "Malformed Trade"



import pytest
import nd2k

from typing import Any


def test_no_input_file(capsys: Any, monkeypatch: Any) -> None:
	monkeypatch.setattr("sys.argv", ["nd2k"])
	with pytest.raises(SystemExit) as exc_info:
		nd2k.main.main()
	assert exc_info.value.code == 1
	assert "Usage: nd2k <novadax-csv>" in capsys.readouterr().out


def test_input_file_does_not_exist(capsys: Any, monkeypatch: Any) -> None:
	monkeypatch.setattr("sys.argv", ["nd2k", "invalid-file.csv"])
	with pytest.raises(SystemExit) as exc_info:
		nd2k.main.main()
	assert exc_info.value.code == 1
	assert "Error: No such file: invalid-file.csv" in capsys.readouterr().out


def test_version_short_form(capsys: Any, monkeypatch: Any) -> None:
	monkeypatch.setattr("sys.argv", ["nd2k", "-v"])
	with pytest.raises(SystemExit) as exc_info:
		nd2k.main.main()
	assert exc_info.value.code == 0
	assert nd2k.__version__ in capsys.readouterr().out


def test_version_long_form(capsys: Any, monkeypatch: Any) -> None:
	monkeypatch.setattr("sys.argv", ["nd2k", "--version"])
	with pytest.raises(SystemExit) as exc_info:
		nd2k.main.main()
	assert exc_info.value.code == 0
	assert nd2k.__version__ in capsys.readouterr().out



import pytest
from datetime import datetime
from decimal import Decimal
from nd2k.converter.organize_rows import (
	parse_date,
	parse_amount,
	parse_trading_pair,
	create_operation,
	create_partial_trade,
)
from nd2k.types import OperationType, PartialTrade, TradingPair, Operation
from ..helpers import fake_op


def test_parse_date() -> None:
	data     = "15/03/1970 23:45:56"
	actual   = parse_date(data)
	expected = datetime(1970, 3, 15, 23, 45, 56)
	assert actual == expected


def test_parse_amount_no_commas() -> None:
	data     = "+12345678901234567890 SHIB(≈R$0)"
	actual   = parse_amount(data)
	expected = Decimal(12345678901234567890)
	assert actual == expected


def test_parse_amount_one_comma() -> None:
	data     = "R$ -355,77"
	actual   = parse_amount(data)
	expected = Decimal("355.77")
	assert actual == expected


def test_parse_amount_multiple_commas() -> None:
	data     = "-121,162,430,769,2304 BABYDOGE2(≈R$0.45)"
	actual   = parse_amount(data)
	expected = Decimal("121162430769.2304")
	assert actual == expected


def test_parse_amount_invalid() -> None:
	data = "No digits present in string"
	with pytest.raises(ValueError) as e:
		parse_amount(data)
	assert str(e.value) == f"No numeric values found in \"{data}\""


def test_parse_trading_pair() -> None:
	data = "Compra(DOGE/BRL)"
	assert parse_trading_pair(data) == TradingPair(base="DOGE", quote="BRL")


def test_parse_trading_pair_invalid() -> None:
	data = "AnythingElse"
	with pytest.raises(ValueError) as e:
		parse_trading_pair(data)
	assert str(e.value) == f"No trading pair found in \"{data}\""


def test_create_operation() -> None:
	csv_line = [
		"15/03/1970 23:45:56",
		"Compra(ABC/BRL)",
		"ABC",
		"-1,234,567 ABC(≈R$87.38)",
		"Sucesso",
	]
	actual   = create_operation(csv_line)
	expected = Operation(
		date    = datetime(1970, 3, 15, 23, 45, 56),
		type    = OperationType.BUY,
		summary = "Compra(ABC/BRL)",
		symbol  = "ABC",
		amount  = Decimal("1234.567"),
		status  = "Sucesso",
	)
	assert actual == expected


def test_create_partial_trade_from_base_asset() -> None:
	base = fake_op(
		summary = "Compra(ABC/XYZ)",
		type    = OperationType.BUY,
		symbol  = "ABC"
	)
	actual   = create_partial_trade(base)
	expected = PartialTrade(
		summary      = "Compra(ABC/XYZ)",
		trading_pair = TradingPair("ABC", "XYZ"),
		base_asset   = base,
	)
	assert actual == expected


def test_create_partial_trade_from_quote_asset() -> None:
	quote = fake_op(
		summary = "Compra(ABC/XYZ)",
		type    = OperationType.BUY,
		symbol  = "XYZ"
	)
	actual   = create_partial_trade(quote)
	expected = PartialTrade(
		summary      = "Compra(ABC/XYZ)",
		trading_pair = TradingPair("ABC", "XYZ"),
		quote_asset  = quote,
	)
	assert actual == expected



from .types import Operation, PartialTrade, Trade


def is_successful(op: Operation) -> bool:
	return op.status == "Sucesso"


def is_part_of_a_trade(op: Operation) -> bool:
	return op.type.name in ["BUY", "SELL", "TRADING_FEE"]


def is_completed(tr: PartialTrade) -> bool:
	return all([
		tr.base_asset,
		tr.quote_asset,
		tr.trading_fee])


def fits_as_base_asset(op: Operation, tr: PartialTrade) -> bool:
	return (
		not tr.base_asset
		and op.summary == tr.summary
		and op.symbol  == tr.trading_pair.base)


def fits_as_quote_asset(op: Operation, tr: PartialTrade) -> bool:
	return (
		not tr.quote_asset
		and op.summary == tr.summary
		and op.symbol  == tr.trading_pair.quote)


def fits_as_trading_fee(op: Operation, tr: PartialTrade) -> bool:
	if tr.trading_fee:
		return False

	if not is_trading_fee(op):
		return False

	if is_a_purchase(tr):
		return op.symbol == tr.trading_pair.base

	if is_a_sale(tr):
		return op.symbol == tr.trading_pair.quote

	raise ValueError("Malformed Trade")


def is_trading_fee(op: Operation) -> bool:
	return op.type.name == "TRADING_FEE"


def is_a_purchase(tr: Trade | PartialTrade) -> bool:
	return tr.type.name == "BUY"


def is_a_sale(tr: Trade | PartialTrade) -> bool:
	return tr.type.name == "SELL"


def is_sending_funds(op: Operation) -> bool:
	return "WITHDRAW" in op.type.name
